-------------------------------- BIGNUM.e -----------------------------------
--             Routines for handling long integer arithmetic               --
--                         J. Craig Gilbert                                --
--                                                                         --
--             Comments, questions, bug reports, suggestions?              --
--               E-mail to cgilbert@cennet.mc.peachnet.edu                 --
--           --------------------------------------------------            --
--  This code is Public Domain, so use it as you wish.  All I ask is that  --
--  you let me know if you find any good optimizations to make.            --
-----------------------------------------------------------------------------
-------------------------- ROUTINE SUMMARY
-- ** SUPPORT PROCEDURES
-- none
--
-- ** SUPPORT FUNCTIONS
-- even_len(bignum s1, bignum s2)
-- fixcarries( bignum n, atom fixtype )
-- trimzero(bignum n)
--
-- ** USER PROCEDURES
-- bn_shownum(bignum n)
--
-- ** USER FUNCTIONS
-- bn_add(bignum num1, bignum num2)
-- bn_approx_sqrt(bignum n)
-- bn_approxroot(bignum n)
-- bn_atomto_other(atom innum, integer tobase)
-- bn_baseconv(bignum innum, integer frombase, integer tobase)
-- bn_compare(bignum n1,bignum n2)
-- bn_curto_atom(bignum innum)
-- bn_dec(bignum innum)
-- bn_diff(bignum n1, bignum n2)
-- bn_div(bignum dvend, bignum dvsor)
-- bn_factor( bignum n )
-- bn_inc(bignum innum)
-- bn_is_even(bignum n)
-- bn_is_one(bignum n)
-- bn_is_zero(bignum n)
-- bn_mult(bignum n1, bignum n2)
-- bn_power(bignum base, bignum exponent)
-- bn_set_base(integer newbase)
-- bn_squarenum(bignum n)
-------------------------- CONVENTIONS USED
-- * Module and global level variables and constants are all caps
--   except for a lower case prefix letter; m for module level and
--   g for global level.  In addition, constants have a secondary
--   prefix, the letter c. Note that this convention is NOT followed
--   in defined types, nor in generic use constants such as YES, NO, etc.
--   And yes, it does cause some constants to look as if McDonald's 
--   sponsored them.  Sue me.
-- * All user functions and procedures (globals) have the prefix 'bn_'.
--------------------------- GENERAL NOTES
-- * 07/09/97 - The maximum number base that can be used as of this writing
--   is base 67,108,864 (gMAXBASE).
-- * 07/09/97 - Signed numbers are not supported at this time.
-- * If you have many consecutive operations to do on a number it is
--   often faster to first convert all numbers to gMAXBASE, peform
--   the operations, then convert back to your base of choice with the
--   final result.  The deciding factor is whether it takes long enough to
--   convert the final result back to base <whatever> to eat up the time
--   saved by performing operations in gMAXBASE.
-- * Calculating roots on low base numbers (binary, etc.) is extremely slow; 
--   avoid it if possible.
-- * If you have *any* ideas or suggestions for improving the speed of these
--   functions I would greatly appreciate them.
-----------------------------------------------------------------------------
-- some constants for convenience, YES/NO, etc.
constant YES = 1,
         NO = 0,
         TRUE = 1

constant mcINTMAX = 9007199254740992,-- largest value Euphoria can represent
                                     -- exactly (I think)
         mcDEFAULTBASE = 10,  -- sets the number base on start-up

         mcADDITION = 1,     -- determines which type of carry operation
         mcSUBTRACTION = -1, -- to correct for in fixcarries()

         mcSAME = 0,     -- use w/ bn_compare()
         mcLEFTBIG = 1,  -- "   "   " 
         mcRIGHTBIG = -1 -- "   "   "

global constant gcMAXBASE = floor(sqrt(mcINTMAX/2)) -- max safe number base

integer  mBASE    -- the current working number base, used by these 
                  -- routines
atom  mSAFEMAX    -- the maximum value allowable at any stage
                  -- of fundamental operations

global integer gCURBASE -- always equals mBASE but is not used by any
                        -- routines in this file; maintained to allow
                        -- the user to check the current working base
                        -- without endangering the integrity of the
                        -- (internally important) variable mBASE


mBASE = mcDEFAULTBASE
gCURBASE = mBASE
mSAFEMAX = mcINTMAX - mBASE*mBASE

sequence mTWO  -- handy in several places to have a 
               -- bignum version of 2 predefined
if mcDEFAULTBASE > 2 then
   mTWO = {2}
else
   mTWO = {1,0}
end if

------------------------ DEFINED TYPES
global type bignum(object n)
    return sequence(n)
end type

global type numbase(object n)
   return integer(n) and n > 1 and n <=gcMAXBASE
end type
------------------------- ROUTINE DECLARATIONS
global procedure bn_shownum(bignum n)
-- displays number; uses separators if the number base is higher than
-- 10
-- each digit is  expressed in base 10, regardless of the current
-- number base setting
integer len
   len = length(n)
   
   if mBASE <= 10 then
      for x = 1 to len by 1 do
         printf(1,"%d",{n[x]})
      end for
   else 
      for x = 1 to (len-1) by 1 do
         printf(1,"%d",{n[x]})
      end for
         printf(1,"%d",{n[len]})
   end if
end procedure


global function bn_set_base(numbase newbase)
-- sets the working number base for all operations; restricts the base to 
-- the range 2 through gcMAXBASE with gcMAXBASE defined as the maximum base 
-- that can be used without causing fundamental operations to produce 
-- numbers greater than the maximum exact value Euphoria can handle.
-- maintains the module variables mBASE, mSAFEMAX and the global
-- gCURBASE, along with the convenience variable mTWO.
-- RETURNS: the previous number base
integer oldbase
   
   if newbase > gcMAXBASE or newbase < 2 then  -- leave base as it is
      oldbase = mBASE                          
      gCURBASE = mBASE                         
   else                         
      oldbase = mBASE
      mBASE = newbase
      gCURBASE = mBASE
      mSAFEMAX = mcINTMAX - mBASE*mBASE
      if mBASE > 2 then
         mTWO = {2}
      else  -- is binary, only legal base in which 2 is not '2'
         mTWO = {1,0}
      end if
   end if

   return oldbase
end function

function even_len(bignum s1, bignum s2)
-- makes two numbers equal in digit length by padding the left side of one
-- with zeroes
-- RETURNS: a two element sequence whose elements are the input numbers
--          in the same order they were passed to the function
integer diff

   diff = length(s1) - length(s2)
   if diff > 0 then
      s2 = repeat(0,diff) & s2
   else
      s1 = repeat(0, diff*(-1)) & s1
   end if
    
   return {s1,s2}
end function


function trimzero(bignum n)
-- removes leading zeros from number
-- RETURNS: the number after zeros have been stripped
integer len, start
   len = length(n)
   if len = 0 then
      return 0
   end if
   
   start = len
   
   for x = 1 to len by 1 do
      if n[x] != 0 then
         start = x
         exit
      end if
   end for
      
   return n[start..len]
end function

global function bn_is_zero(bignum n)
-- determines whether the input number is zero
-- RETURNS: 1 if the number is zero, 0 if not
integer len

    len = length(n)
    if n[len] = 0 then
        for x = (len-1) to 1 by -1 do
            if n[x] != 0 then
                return NO
            end if    
        end for
    else
        return NO
    end if
    

return YES
end function
                                             
function bn_is_one(bignum n)
-- determines whether the input number is one
-- RETURNS: 1 if the number is one, 0 if not
integer len

    len = length(n)
    if n[len] = 1 then
        for x = (len-1) to 1 by -1 do
            if n[x] != 0 then
                return NO
            end if    
        end for
    else
        return NO
    end if
   
return YES
end function

function bn_is_even(bignum n)
-- determines if the input number is even
-- RETURNS: 1 if the number is even, 0 if not
integer retval, len
   
   len = length(n)
   if len < 1 then
      retval = YES
   else
      if and_bits(n[len],1) = 0 then   -- if number is not odd
         retval = YES
      else
         retval = NO
      end if
   end if

   return retval
end function


global function bn_compare(bignum n1,bignum n2)
-- compares two numbers
-- RETURNS: -1 if n1 < n2
--           0 if n1 = n2
--           1 if n1 > n2
bignum bucket, even1, even2
integer compval
      
    bucket = even_len(n1,n2)
    even1 = bucket[1]
    even2 = bucket[2]
    compval = compare(even1,even2)
   
   return compval
end function

function fixcarries( bignum n, atom fixtype )
-- adjusts the numbers passed in so that no 'digit' is greater
-- than the maximum allowable digit value for the current number base, 
-- or less than zero (as in mcSUBTRACTION)
-- RETURNS: the adjusted number
bignum result
integer redo, len
atom curdigit, numcarries, numbases
result = n
redo = 1
if fixtype = mcADDITION then    
   while redo = YES do
      result = 0 & result
      len = length(result)
      redo = NO
      for x = len to 2 by -1 do
         curdigit = result[x] 
         if curdigit > (mBASE-1) then -- carry the excess over
            numcarries = floor(curdigit/mBASE)
            if numcarries > mSAFEMAX then -- keep the carryover from
               numcarries = mSAFEMAX      -- exceeding the max atom
               redo = YES                 -- value
            end if
            result[x] = curdigit - mBASE*(numcarries)
            result[x-1] = result[x-1] + numcarries
         end if
      end for
   end while

   redo = NO
   while result[1] > (mBASE-1) or redo = 1 do -- fix most significant
      result = 0 & result                           -- digit
      numcarries = floor(result[2]/mBASE)
      if numcarries > mSAFEMAX then
         numcarries = mSAFEMAX
         redo = YES
      end if
      result[2] = result[2] - mBASE*(numcarries)
      result[1] = result[1] + numcarries
   end while

else -- fix for mcSUBTRACTION, essentially adjusting for the borrows
     -- that would have been made if subtracting longhand
   while redo = YES do
      result = 0 & result
      len = length(result)
      redo = NO
      for x = len to 2 by -1 do
         curdigit = result[x] 
         if curdigit < 0 then   -- bring in the next digit's excess
            numbases = curdigit/(-mBASE)
            if numbases != 1 then
               numcarries = floor(numbases) + 1
            else
               numcarries = 1
            end if
            if numcarries > mSAFEMAX then -- keep the carryover from
               numcarries = mSAFEMAX      -- exceeding the max atom
               redo = YES                 -- value
            end if
            result[x] = curdigit + mBASE*(numcarries)
            result[x-1] = result[x-1] - numcarries
         end if
      end for
   end while
end if
    
return trimzero(result)
end function   
   
   
global function bn_add(bignum n1, bignum n2)
-- add two numbers
-- RETURNS: the sum of the numbers
bignum sum, bucket

    bucket = even_len(n1,n2)
    sum = bucket[1] + bucket[2]
    sum = fixcarries( sum, mcADDITION )
    
    return sum
end function

global function bn_inc(bignum n)
-- add one to the number
-- RETURNS: the incremented number
integer len

   len = length(n)
   n[len] = n[len] + 1
   if (n[len] > (mBASE-1)) then
      n = fixcarries( n, mcADDITION)
   end if
   return n
end function


global function bn_diff(bignum n1, bignum n2)
-- calculate the POSITIVE difference between two numbers
-- RETURNS: the difference
bignum diff, bucket

    if bn_compare(n1,n2) = mcLEFTBIG then
        bucket = even_len(n1,n2)
        diff = bucket[1] - bucket[2]
    else
        bucket = even_len(n1,n2)
        diff = bucket[2] - bucket[1]
    end if

    diff = fixcarries( diff, mcSUBTRACTION )

   return diff
end function

global function bn_dec(bignum n)
-- subtract one from the number
-- RETURNS: the decremented number
integer len

   len = length(n)
   n[len] = n[len] - 1
   if n[len] < 0 then 
      n = fixcarries( n, mcSUBTRACTION )
   end if
   return n
end function

global function bn_mult(bignum n1, bignum n2)
-- multiplies two numbers by brute force method
-- RETURNS: the product of the numbers
bignum product, curLine
integer len

    len = length(n1)
    product = {0}
    for x = len to 1 by -1 do
        curLine = n2*n1[x] & repeat(0,len-x)
        product = bn_add(product,curLine)
        if x = 1 then
            product = fixcarries(product, mcADDITION)         
        end if        
    end for

    return product
end function

global function bn_squarenum(bignum n)
-- RETURNS: the square of the number
   return bn_mult(n,n)
end function

global function bn_curto_atom(bignum innum)
-- convert a number in bignum form to an atom, base 10, using Horner's
-- method for evaluating polynomials
-- Note: does not check to see if the number is actually within
-- the limits of exact representation
-- RETURNS: base 10 representation of the number
integer fromLen
atom outnum

   fromLen = length(innum)
    outnum = innum[1]
    for curdigit = 2 to fromLen by 1 do
        outnum = outnum*mBASE + innum[curdigit]
    end for
      
   return outnum
end function

global function bn_atomto_other(atom innum, numbase tobase)
-- convert an atom (base 10) number to another base and represent it
-- in long integer form (bignum type)
-- RETURNS: the number in long integer form as a 'tobase' number
bignum outnum
integer redo
atom numcarries

   if tobase > gcMAXBASE then
       outnum = {-1}
   elsif innum < tobase then
      outnum = {innum}
   else
      outnum = {innum}
      redo = YES
      while redo = YES do
         redo = NO
         outnum = 0 & outnum
         for x = length(outnum) to 2 by -1 do
            numcarries = floor(outnum[x]/tobase)
            outnum[x] = outnum[x] - tobase*numcarries
            outnum[x-1] = outnum[x-1] + numcarries
         end for
         if outnum[1] >= tobase then
            redo = YES
         end if
      end while
   end if

   return outnum
end function

global function bn_div(bignum divend, bignum dvsor)
-- divide two numbers by 'longhand' method
-- RETURNS: two-element sequence containing quotient and remainder
-- { {Quotient}, {Remainder} }
bignum curdividend, trialQ, trialSub, Q, R, gap
sequence bucket
integer dvsorlen, divendlen, next, up, down, direction, curdivendlen, trialmatch
atom dvsorFirst, offset

    Q = {}
    R = {}
    
    dvsorlen = length(dvsor)
    divendlen = length(divend)
    
    if divendlen < dvsorlen then
        return { {0}, divend }
    end if
    
    if dvsorlen > 1 then
        curdividend = divend[1..dvsorlen-1]
        next = dvsorlen 
    else
        curdividend = {}
        next = 1
    end if

   curdivendlen = length(curdividend)
   dvsorFirst = dvsor[1]
   up = 1
   down = -1
       
    while next <= divendlen do   -- get quotient digits
        curdividend = curdividend & divend[next]
        next = next + 1
        curdivendlen = curdivendlen + 1
        if curdivendlen = dvsorlen then
            trialQ = {floor(curdividend[1]/dvsorFirst)}
        else
            trialQ = {floor( bn_curto_atom(curdividend[1..2])/dvsorFirst)}
            if trialQ[1] >= mBASE then
                trialQ[1] = mBASE - 1
            end if
        end if
        trialSub = bn_mult( dvsor, trialQ )
        gap = bn_diff(trialSub, curdividend) 
        if bn_compare( gap, dvsor ) = mcLEFTBIG then
            bucket = bn_div( gap, dvsor )
        else
            bucket = {{0},{0}}
        end if       
        if length(bucket[1]) = 1 then
            offset = bucket[1][1]
        else 
            offset = mBASE-1
        end if
      
        direction = up

        while offset > 1 do   -- zero in on the current quotient
            while bn_compare({offset},trialQ) = mcLEFTBIG do
                offset = floor(offset/2)
            end while

            if direction = up then
                trialQ = bn_add(trialQ,{offset})
            else
                trialQ = bn_diff(trialQ,{offset})
            end if

            trialSub = bn_mult( trialQ, dvsor )
            trialmatch = bn_compare( trialSub, curdividend)
            if trialmatch = direction then
                offset = floor(offset/2)
                direction = direction * (-1)
            elsif trialmatch = 0 then
               exit
            end if                
        end while

        -- final adjustments to current quotient
        while bn_compare( trialSub, curdividend ) = mcRIGHTBIG do
            trialQ = bn_inc(trialQ)
            trialSub = bn_mult( trialQ, dvsor )
        end while
        while bn_compare( trialSub, curdividend ) = mcLEFTBIG do
            trialQ = bn_dec(trialQ)
            trialSub = bn_mult( trialQ, dvsor )
        end while

        Q = Q & trialQ
        curdividend = bn_diff(curdividend,trialSub)
    end while

    R = curdividend

    return {trimzero(Q), trimzero(R)}    
end function

global function bn_power(bignum base, bignum exponent)
-- raise number(base) to power(exponent)
-- uses squaring and recursion to lower the number of multiplications,
-- like so: x^37 = x^32 * x^4 * x^1
--               = (((((x^2)^2)^2)^2)^2) * ((x^2)^2) * x^1 
-- RETURNS: base^exponent
bignum result, expdiff, nextround

   if bn_is_zero(exponent) then
      result = {1}
      return result
   elsif bn_is_one(exponent) then
      result = base
      return result
   elsif compare(exponent,mTWO) = mcSAME then
      result = bn_squarenum(base)
      return result
   else  -- handle powers greater than two
      result = bn_squarenum(base)
      exponent = bn_diff(exponent,mTWO)  
      expdiff = mTWO
      
      -- square while the accumulated power of the result
      -- is less than the left over power of the target
      while bn_compare(exponent, expdiff) = mcLEFTBIG do
         result = bn_squarenum(result)
         exponent = bn_diff(exponent,expdiff)
         expdiff = bn_mult(expdiff,mTWO)
      end while   
      if bn_compare(exponent, expdiff) = mcSAME then
         result = bn_squarenum(result)
         return result
      end if   
        
    nextround = bn_power(base,exponent)

    result = bn_mult(result, nextround)
    
   end if

   return result
end function
                
global function bn_baseconv(bignum innum, numbase frombase, numbase tobase)
-- convert a number from current base to the specified new number base;
-- uses Horner's method for evaluating polynomials, which seems to be
-- significantly faster than base conversion by iterative dividing
-- RETURNS: the original number in the new number base
bignum outnum, frombase_newform
integer fromLen
atom oldbase

   oldbase = bn_set_base(tobase)
   fromLen = length(innum)
-- from here on, all work is done in the target number base, treating the
-- original number as a polynomial evaluated at x = frombase.
-- For instance: to convert 4560000000 to base 8 (octal):
-- 4560000000 in base 10(bignum) = {4,5,6,0,0,0,0,0,0,0}
-- represent as polynomial:{4}*{1,0}^{1,0} + {5}*{1,0}^{9} + {6}*{1,0}^{8} 
-- convert parts to octal:{4}*{1,2}^{1,2} + {5}*{1,2}^{1,1} + {6}*{1,2}^{1,0}
-- evaluate the polynomial (in octal) and you have the octal result:
-- {4,1,7,6,3,0,1,2,0,0,0}
  frombase_newform = bn_atomto_other(frombase,tobase)

  outnum = bn_atomto_other( innum[1], tobase )
  for curdigit = 2 to fromLen by 1 do
      outnum = bn_mult(outnum, frombase_newform)
      outnum = bn_add(outnum, bn_atomto_other( innum[curdigit], tobase ) )
  end for

  oldbase = bn_set_base(oldbase)
  return outnum
end function

global function bn_approx_sqrt(bignum n)
-- find the square root of number, leaving off decimals
-- uses 'longhand' method of finding square roots
-- RETURNS: closest integer less than or equal to the square root 
--          of the number
bignum curquot, curgoal, curroot, answer, lastgoal, curdivisor
bignum lastresult, trialSub
integer numdigits, len, div, curslot,lendivisor,last, up, down, digitmatch
atom offset, curBase10goal


    up = 1
    down = -1
    
    len = length(n)
    if len < 1 then
        return {-1}
    end if
   
    if and_bits(len,1) = 1 then   -- if not even
        numdigits = floor(len/2) + 1
        div = 1
    else
        numdigits = len/2
        div = 2
    end if

    answer = repeat(0,numdigits)
   
    -- get first digit; at this point you want the integer square root
    --   of curgoal
    curgoal = n[1..div]
    curBase10goal = bn_curto_atom(curgoal)
    curroot = {floor(sqrt(curBase10goal))}
    answer[1] = curroot[1]
   
    -- get rest of the root; from here on you want the number which
    -- is the quotient of curgoal divided by curdivisor with this
    -- quotient also being the last digit of the curdivisor
    -- i.e.  270 divided by 4?  , choose 5 as the missing digit
    --       and you have 270/45 = 5; even though the proper quotient
    --       is 6, this is the closest one which allows you to match 
    --       the quotient and the last digit of the divisor
    -- after each round, the next set of two digits is brought down and
    -- added to the end of curgoal;  the curdivisor is doubled and
    -- multiplied by mBASE, and the psuedo-quotient search starts again
    lastgoal = curgoal
    lastresult = bn_power(curroot,{2})
    curslot = 1
    while div < len  do
        curgoal = bn_diff(lastgoal, lastresult) & n[ (div+1) .. (div+2) ]
        curquot = answer[1..curslot]
        curdivisor = bn_mult(curquot, {2}) & 0
        curslot = curslot + 1
        div = div + 2
         
        lendivisor = length(curdivisor)
        digitmatch = 1
        last = up
        offset = floor((mBASE-1)/2)
        while offset > 1 do   -- zero in on quotient that matches last digit

            trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
            while bn_compare( trialSub, curgoal ) = mcRIGHTBIG do
                curdivisor[lendivisor] = curdivisor[lendivisor] + offset
                if curdivisor[lendivisor] >= mBASE then
                    curdivisor[lendivisor] = mBASE-1
                    exit
                end if
                trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
            end while
            
            offset = floor(offset/2)
            if offset < 2 then
                exit
            end if
            
            trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
            while bn_compare( trialSub, curgoal ) = mcLEFTBIG do
                curdivisor[lendivisor] = curdivisor[lendivisor] - offset
                if curdivisor[lendivisor] < 0 then
                    curdivisor[lendivisor] = 0
                    exit
                end if
                trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
            end while
            
        end while        

        trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
        while bn_compare( trialSub, curgoal ) = mcRIGHTBIG do
            curdivisor[lendivisor] = curdivisor[lendivisor] + 1
            if curdivisor[lendivisor] >= mBASE then
                curdivisor[lendivisor] = mBASE-1
                exit
            end if
            trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
        end while
            
        trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
        while bn_compare( trialSub, curgoal ) = mcLEFTBIG do
            curdivisor[lendivisor] = curdivisor[lendivisor] - 1
            if curdivisor[lendivisor] < 0 then
                curdivisor[lendivisor] = 0
                exit
            end if
            trialSub = bn_mult(curdivisor,{curdivisor[lendivisor]})
        end while
      
        answer[curslot] = curdivisor[lendivisor]
      
        lastgoal = curgoal
        lastresult = bn_mult(curdivisor,{curdivisor[lendivisor]})
   end while
   
   return answer
end function


global function bn_approxroot(bignum n, atom rootnum)
-- find integer root of number by essentially trial and error, 
-- although with a half-arse sort of binary search
-- Note: because of the integer return, extremely high rootnums will
--       result in roots = 1.
-- RETURNS: closest integer less than or equal to the 'rootnum'th root 
--          the of number 'n'
bignum root, offset, pow, tempgoal
sequence bucket
integer numdigits, adjust, div

    if rootnum = 2 then
        return bn_approx_sqrt(n)
    end if
   
   -- First, work up a guess that is at least in the neighborhood
    numdigits = floor(length(n)/rootnum)
    adjust = length(n) - numdigits*rootnum
    if adjust > 0 then
        div = adjust
        numdigits = numdigits + 1
    else
        div = rootnum
    end if
    
    root = repeat(0,numdigits)
    
    tempgoal = n[1..div]
    if length(tempgoal) > 2 and div != length(tempgoal) then
        root[1] = bn_curto_atom( bn_approxroot(tempgoal,rootnum) )
    else
        root[1] = floor(power(bn_curto_atom( tempgoal ), 1/rootnum) )
    end if

   -- Second, find the actual root by searching up and down from
   -- the original (and subsequent) guess(es)
    pow = bn_atomto_other( rootnum, mBASE )
    bucket = bn_div( root, mTWO )
    offset = bucket[1]
    while not bn_is_zero(offset) do    
        while bn_compare( bn_power(root, pow), n ) = mcRIGHTBIG do
            root = bn_add(root, offset)
        end while
        bucket = bn_div( offset, mTWO )
        offset = bucket[1]
        if bn_is_zero(offset) then
            exit
        end if
        while bn_compare( bn_power(root, pow), n ) = mcLEFTBIG do
            root = bn_diff(root, offset)
        end while
        
    end while
    
    while bn_compare( bn_power(root, pow), n ) = mcRIGHTBIG do
        root = bn_inc(root)        
    end while
    while bn_compare( bn_power(root, pow), n ) = mcLEFTBIG do
        root = bn_dec(root)
    end while

    return root
end function

global function bn_factor(bignum n)
-- get the prime factors of a number and the powers to which they must
-- be raised to rebuild the number
-- Note: function is immensely slow, especially if it hits a large prime
-- RETURNS: all prime factors of number 'n' in the form:
--           { {factor}, {power}, {factor}, {power} , . . .}
bignum divend, dvsor, pow
sequence bucket, factors
   
   divend = n
   dvsor = mTWO
   factors = {}
   while bn_compare(dvsor, divend) = mcRIGHTBIG do
      pow = {0}
      while TRUE do
         bucket = bn_div( divend, dvsor )
         if not bn_is_zero(bucket[2]) then
            exit
         end if
         divend = bucket[1]
         pow = bn_inc(pow)
      end while

      if not bn_is_zero(pow) then
         factors = append(factors,dvsor)
         factors = append(factors,pow)
      end if

      dvsor = bn_inc(dvsor)
      while bn_is_even(dvsor) do
         dvsor = bn_inc(dvsor)
      end while
   end while

   pow = {1}
   factors = append(factors,divend)
   factors = append(factors,pow)

   if bn_is_zero(factors[2]) then
      factors = factors[3..length(factors)]
   end if
   
   return factors
end function

